'use strict';

var path = require('path');
var fs = require('fs');
var package_json = require('tslib/package.json');
var process = require('process');
var typescript = require('typescript');

function createServiceHost(options, filenames, cwd) {
	const normalizePath = (path$1) => path.resolve(path.normalize(path$1));
	const moduleResolutionHost = createModuleResolutionHost();
	const files = {}; // normalized filename => {version, snap, text}
	filenames.forEach(filename => files[normalizePath(filename)] = null);

	return {
		getDirectories: typescript.sys.getDirectories,
		directoryExists: typescript.sys.directoryExists,
		readDirectory: typescript.sys.readDirectory,
		getDefaultLibFileName: typescript.getDefaultLibFilePath,

		fileExists(filename) {
			filename = normalizePath(filename);
			return filename in files || typescript.sys.fileExists(filename);
		},

		readFile(filename) {
			return fs.readFileSync(normalizePath(filename), "utf-8");
		},

		getCompilationSettings() {
			return options;
		},

		getCurrentDirectory() {
			return cwd;
		},

		getScriptFileNames() {
			return Object.keys(files);
		},

		getScriptVersion(filename) {
			const f = files[normalizePath(filename)];
			return f ? f.version.toString() : "";
		},

		getScriptSnapshot(filename) {
			let f = files[normalizePath(filename)];
			if(!f) {
				f = this.addFile(filename, this.readFile(filename));
			}
			return f.snap;
		},

		resolveModuleNames(moduleNames, containingFile) {
			return moduleNames.map(name => this.resolveModuleName(name, containingFile));
		},

		getNewLine() {
			return options.newLine || typescript.sys.newLine;
		},

		// additional methods
		containsFile(filename) {
			return normalizePath(filename) in files;
		},

		resolveModuleName(moduleName, containingFile) {
			const {resolvedModule} = typescript.resolveModuleName(moduleName, containingFile, options, moduleResolutionHost);
			if(resolvedModule) {
				resolvedModule.resolvedFileName = normalizePath(resolvedModule.resolvedFileName);
				resolvedModule.originalFileName = resolvedModule.resolvedFileName;
			}
			return resolvedModule;
		},

		addFile(filename, text) {
			filename = normalizePath(filename);

			const snap = typescript.ScriptSnapshot.fromString(text);
			snap.getChangeRange = () => {};

			let file = files[filename];
			if(!file) {
				file = {version: 0};
				files[filename] = file;
			}
			++file.version;
			file.snap = snap;
			file.text = text;
			return file;
		},
	};
}


function createModuleResolutionHost() {
	return {
		fileExists(filename) {
			return fs.existsSync(filename);
		},

		readFile(filename) {
			return fs.readFileSync(filename, "utf-8")
		},
	};
}

function fileFilter(tsconfig) {
	const files = fileMatcher(tsconfig.files || []);
	const glob = globMatcher(tsconfig.exclude || [], tsconfig.include || []);

	return (filename) => {
		filename = path.resolve(filename);
		return files(filename) || glob(filename);
	};
}


function fileMatcher(files) {
	files = files.map(path.resolve);
	return (filename) => files.includes(filename);
}

function globMatcher(excludePatterns, includePatterns) {
	const exclude = excludePatterns.map(buildRegexp);
	const include = includePatterns.map(buildRegexp);

	return (filename) => {
		for(let i = 0; i < exclude.length; ++i) {
			if(exclude[i].test(filename)) {
				return false;
			}
		}

		for(let i = 0; i < include.length; ++i) {
			if(include[i].test(filename)) {
				return true;
			}
		}

		return false;
	};
}

function buildRegexp(pattern) {
	const parts = pattern.replace("\\", "/").split("/").map(part => {
		if(part === "**") {
			return "([^/]*(/|$))*";
		}

		let rx = "";
		for(let i = 0; i < part.length; ++i) {
			switch(part[i]) {
			case "*":
				rx += "[^/]*";
				break;

			case "?":
				rx += "[^/]";
				break;

			default:
				rx += part[i];
				break;
			}
		}
		return rx;
	});

	return new RegExp(path.resolve(parts.join("/")));
}

const defaultCompilerOptions = {
	module: typescript.ModuleKind.ES2015,
	moduleResolution: typescript.ModuleResolutionKind.NodeJs,
};


function createService(tsconfig) {
	const cwd = process.cwd();
	const {options, fileNames, errors} = typescript.parseJsonConfigFileContent(tsconfig, typescript.sys, path.dirname(""), defaultCompilerOptions);
	if(errors.length) {
		throw new Error(errorMessage(errors[0], cwd));
	}

	Object.assign(options, {
		target: typescript.ScriptTarget.ES2015,
		noEmitOnError: false,
		suppressOutputPathCheck: true,
		allowNonTsExtensions: true,
	});

	const host = createServiceHost(options, fileNames, cwd);
	const reg = typescript.createDocumentRegistry();
	const svc = typescript.createLanguageService(host, reg);

	const filter = fileFilter(tsconfig);

	return {
		emit(filename, code) {
			host.addFile(filename, code);

			const output = svc.getEmitOutput(filename);
			let diag = svc.getSyntacticDiagnostics(filename);
			if(!diag.length) {
				diag = svc.getCompilerOptionsDiagnostics();
				if(!diag.length) {
					diag = svc.getSemanticDiagnostics(filename);
				}
			}

			output.errors = [];
			output.warnings = [];
			diag.forEach(d => {
				const msg = errorMessage(d, cwd);
				switch(d.category) {
				case typescript.DiagnosticCategory.Error:
					output.errors.push(msg);
					break;
				case typescript.DiagnosticCategory.Warning:
					output.warnings.push(msg);
					break;
				}
			});
			return output;
		},

		filter(filename) {
			return filter(filename) || host.containsFile(filename);
		},

		resolveModuleName(importee, importer) {
			const {resolvedFileName} = host.resolveModuleName(importee, importer) || {};
			return resolvedFileName;
		},
	};
}


function errorMessage(diagnostic, cwd) {
	const text = typescript.flattenDiagnosticMessageText(diagnostic.messageText, "\n");
	if(!diagnostic.file) {
		return `tsc: ${text}`;
	}

	const {line} = diagnostic.file.getLineAndCharacterOfPosition(diagnostic.start);
	const file = path.relative(cwd, diagnostic.file.fileName);
	return `${file}:${line+1}: ${text}`;
}

function tsc(tsconfig) {
	tsconfig = tsconfig || {};
	tsconfig.compilerOptions = tsconfig.compilerOptions || {};

	const loadTslib = tslibLoader();
	const service = createService(tsconfig);

	let bundleDecls = [];

	return {
		name: "tsc",

		options(opts) {
			if(tsconfig.sourceMap == null) {
				tsconfig.sourceMap = opts.sourcemap;
			}
		},

		resolveId(importee, importer) {
			if(importee === package_json.name) {
				return `\0${package_json.name}`;
			} else if(!importer) {
				return null;
			}
			importer = importer.replace("\\", "/");

			const resolvedFileName = service.resolveModuleName(importee, importer);
			return resolvedFileName && !resolvedFileName.endsWith(".d.ts")
				? resolvedFileName
				: null;
		},

		load(id) {
			if(id === `\0${package_json.name}`) {
				return loadTslib();
			}
			return null;
		},

		transform(code, id) {
			if(!service.filter(id)) {
				return null;
			}

			const {outputFiles, errors, warnings} = service.emit(id, code);
			if(warnings && warnings.length) {
				warnings.forEach(this.warn);
			}
			if(errors && errors.length) {
				errors.forEach(err => console.error(`ERROR: ${err}`));
				this.error("typescript compilation failed");
			}

			const js = outputFiles.find(f => f.name.endsWith(".js") || f.name.endsWith(".jsx"));
			const map = outputFiles.find(f => f.name.endsWith(".map"));
			const dts = outputFiles.find(f => f.name.endsWith(".d.ts"));
			if(dts) {
				bundleDecls.push(dts);
			}

			return !js ? null : {
				code: js.text,
				map: map ? JSON.parse(map.text) : {mappings: ""},
			};
		},

		generateBundle(opts, bundle, isWrite) {
			if(!isWrite || !bundleDecls.length) {
				return;
			}

			const dir = tsconfig.compilerOptions.declarationDir || path.dirname(opts.file);
			return mkdirAll(dir).then(() => Promise.all(bundleDecls.map(decl => {
				const file = path.basename(decl.name);
				return emitFile(path.join(dir, file), decl.text);
			})));
		},
	};
}

function tslibLoader() {
	let src;
	return () => {
		if(src) {
			return Promise.resolve(src);
		}

		return new Promise((resolve, reject) => {
			fs.readFile(require.resolve(`${package_json.name}/${package_json.module}`), "utf8", (err, data) => {
				if(err) {
					reject(err);
				} else {
					resolve(src = data);
				}
			});
		});
	}
}

function mkdirAll(p) {
	return new Promise((resolve, reject) => {
		fs.mkdir(p, (err) => {
			if(!err) {
				return resolve();
			}

			if(err.code === "ENOENT") {
				return mkdirAll(path.dirname(p)).then(() => mkdirAll(p));
			}

			fs.stat(p, (e, s) => {
				if(e) {
					reject(e);
				} else if(!s.isDirectory()) {
					reject(err);
				} else {
					resolve();
				}
			});
		});
	});
}

function emitFile(filename, data) {
	return new Promise((resolve, reject) => {
		fs.writeFile(filename, data, err => {
			if(err) {
				reject(err);
			} else {
				resolve();
			}
		});
	});
}

module.exports = tsc;
//# sourceMappingURL=rollup-plugin-tsc.cjs.js.map
