import { resolve, normalize, dirname, relative, basename, join } from 'path';
import { readFileSync, existsSync, mkdir, stat, writeFile, readFile } from 'fs';
import { name, module } from 'tslib/package.json';
import { cwd } from 'process';
import { sys, getDefaultLibFilePath, resolveModuleName, ScriptSnapshot, DiagnosticCategory, ModuleKind, ModuleResolutionKind, parseJsonConfigFileContent, ScriptTarget, createDocumentRegistry, createLanguageService, flattenDiagnosticMessageText } from 'typescript';

function createServiceHost(options, filenames, cwd) {
	const normalizePath = (path) => resolve(normalize(path));
	const moduleResolutionHost = createModuleResolutionHost();
	const files = {}; // normalized filename => {version, snap, text}
	filenames.forEach(filename => files[normalizePath(filename)] = null);

	return {
		getDirectories: sys.getDirectories,
		directoryExists: sys.directoryExists,
		readDirectory: sys.readDirectory,
		getDefaultLibFileName: getDefaultLibFilePath,

		fileExists(filename) {
			filename = normalizePath(filename);
			return filename in files || sys.fileExists(filename);
		},

		readFile(filename) {
			return readFileSync(normalizePath(filename), "utf-8");
		},

		getCompilationSettings() {
			return options;
		},

		getCurrentDirectory() {
			return cwd;
		},

		getScriptFileNames() {
			return Object.keys(files);
		},

		getScriptVersion(filename) {
			const f = files[normalizePath(filename)];
			return f ? f.version.toString() : "";
		},

		getScriptSnapshot(filename) {
			let f = files[normalizePath(filename)];
			if(!f) {
				f = this.addFile(filename, this.readFile(filename));
			}
			return f.snap;
		},

		resolveModuleNames(moduleNames, containingFile) {
			return moduleNames.map(name => this.resolveModuleName(name, containingFile));
		},

		getNewLine() {
			return options.newLine || sys.newLine;
		},

		// additional methods
		containsFile(filename) {
			return normalizePath(filename) in files;
		},

		resolveModuleName(moduleName, containingFile) {
			const {resolvedModule} = resolveModuleName(moduleName, containingFile, options, moduleResolutionHost);
			if(resolvedModule) {
				resolvedModule.resolvedFileName = normalizePath(resolvedModule.resolvedFileName);
				resolvedModule.originalFileName = resolvedModule.resolvedFileName;
			}
			return resolvedModule;
		},

		addFile(filename, text) {
			filename = normalizePath(filename);

			const snap = ScriptSnapshot.fromString(text);
			snap.getChangeRange = () => {};

			let file = files[filename];
			if(!file) {
				file = {version: 0};
				files[filename] = file;
			}
			++file.version;
			file.snap = snap;
			file.text = text;
			return file;
		},
	};
}


function createModuleResolutionHost() {
	return {
		fileExists(filename) {
			return existsSync(filename);
		},

		readFile(filename) {
			return readFileSync(filename, "utf-8")
		},
	};
}

function fileFilter(tsconfig) {
	const files = fileMatcher(tsconfig.files || []);
	const glob = globMatcher(tsconfig.exclude || [], tsconfig.include || []);

	return (filename) => {
		filename = resolve(filename);
		return files(filename) || glob(filename);
	};
}


function fileMatcher(files) {
	files = files.map(resolve);
	return (filename) => files.includes(filename);
}

function globMatcher(excludePatterns, includePatterns) {
	const exclude = excludePatterns.map(buildRegexp);
	const include = includePatterns.map(buildRegexp);

	return (filename) => {
		for(let i = 0; i < exclude.length; ++i) {
			if(exclude[i].test(filename)) {
				return false;
			}
		}

		for(let i = 0; i < include.length; ++i) {
			if(include[i].test(filename)) {
				return true;
			}
		}

		return false;
	};
}

function buildRegexp(pattern) {
	const parts = pattern.replace("\\", "/").split("/").map(part => {
		if(part === "**") {
			return "([^/]*(/|$))*";
		}

		let rx = "";
		for(let i = 0; i < part.length; ++i) {
			switch(part[i]) {
			case "*":
				rx += "[^/]*";
				break;

			case "?":
				rx += "[^/]";
				break;

			default:
				rx += part[i];
				break;
			}
		}
		return rx;
	});

	return new RegExp(resolve(parts.join("/")));
}

const defaultCompilerOptions = {
	module: ModuleKind.ES2015,
	moduleResolution: ModuleResolutionKind.NodeJs,
};


function createService(tsconfig) {
	const cwd$1 = cwd();
	const {options, fileNames, errors} = parseJsonConfigFileContent(tsconfig, sys, dirname(""), defaultCompilerOptions);
	if(errors.length) {
		throw new Error(errorMessage(errors[0], cwd$1));
	}

	Object.assign(options, {
		target: ScriptTarget.ES2015,
		noEmitOnError: false,
		suppressOutputPathCheck: true,
		allowNonTsExtensions: true,
	});

	const host = createServiceHost(options, fileNames, cwd$1);
	const reg = createDocumentRegistry();
	const svc = createLanguageService(host, reg);

	const filter = fileFilter(tsconfig);

	return {
		emit(filename, code) {
			host.addFile(filename, code);

			const output = svc.getEmitOutput(filename);
			let diag = svc.getSyntacticDiagnostics(filename);
			if(!diag.length) {
				diag = svc.getCompilerOptionsDiagnostics();
				if(!diag.length) {
					diag = svc.getSemanticDiagnostics(filename);
				}
			}

			output.errors = [];
			output.warnings = [];
			diag.forEach(d => {
				const msg = errorMessage(d, cwd$1);
				switch(d.category) {
				case DiagnosticCategory.Error:
					output.errors.push(msg);
					break;
				case DiagnosticCategory.Warning:
					output.warnings.push(msg);
					break;
				}
			});
			return output;
		},

		filter(filename) {
			return filter(filename) || host.containsFile(filename);
		},

		resolveModuleName(importee, importer) {
			const {resolvedFileName} = host.resolveModuleName(importee, importer) || {};
			return resolvedFileName;
		},
	};
}


function errorMessage(diagnostic, cwd) {
	const text = flattenDiagnosticMessageText(diagnostic.messageText, "\n");
	if(!diagnostic.file) {
		return `tsc: ${text}`;
	}

	const {line} = diagnostic.file.getLineAndCharacterOfPosition(diagnostic.start);
	const file = relative(cwd, diagnostic.file.fileName);
	return `${file}:${line+1}: ${text}`;
}

function tsc(tsconfig) {
	tsconfig = tsconfig || {};
	tsconfig.compilerOptions = tsconfig.compilerOptions || {};

	const loadTslib = tslibLoader();
	const service = createService(tsconfig);

	let bundleDecls = [];

	return {
		name: "tsc",

		options(opts) {
			if(tsconfig.sourceMap == null) {
				tsconfig.sourceMap = opts.sourcemap;
			}
		},

		resolveId(importee, importer) {
			if(importee === name) {
				return `\0${name}`;
			} else if(!importer) {
				return null;
			}
			importer = importer.replace("\\", "/");

			const resolvedFileName = service.resolveModuleName(importee, importer);
			return resolvedFileName && !resolvedFileName.endsWith(".d.ts")
				? resolvedFileName
				: null;
		},

		load(id) {
			if(id === `\0${name}`) {
				return loadTslib();
			}
			return null;
		},

		transform(code, id) {
			if(!service.filter(id)) {
				return null;
			}

			const {outputFiles, errors, warnings} = service.emit(id, code);
			if(warnings && warnings.length) {
				warnings.forEach(this.warn);
			}
			if(errors && errors.length) {
				errors.forEach(err => console.error(`ERROR: ${err}`));
				this.error("typescript compilation failed");
			}

			const js = outputFiles.find(f => f.name.endsWith(".js") || f.name.endsWith(".jsx"));
			const map = outputFiles.find(f => f.name.endsWith(".map"));
			const dts = outputFiles.find(f => f.name.endsWith(".d.ts"));
			if(dts) {
				bundleDecls.push(dts);
			}

			return !js ? null : {
				code: js.text,
				map: map ? JSON.parse(map.text) : {mappings: ""},
			};
		},

		generateBundle(opts, bundle, isWrite) {
			if(!isWrite || !bundleDecls.length) {
				return;
			}

			const dir = tsconfig.compilerOptions.declarationDir || dirname(opts.file);
			return mkdirAll(dir).then(() => Promise.all(bundleDecls.map(decl => {
				const file = basename(decl.name);
				return emitFile(join(dir, file), decl.text);
			})));
		},
	};
}

function tslibLoader() {
	let src;
	return () => {
		if(src) {
			return Promise.resolve(src);
		}

		return new Promise((resolve, reject) => {
			readFile(require.resolve(`${name}/${module}`), "utf8", (err, data) => {
				if(err) {
					reject(err);
				} else {
					resolve(src = data);
				}
			});
		});
	}
}

function mkdirAll(p) {
	return new Promise((resolve, reject) => {
		mkdir(p, (err) => {
			if(!err) {
				return resolve();
			}

			if(err.code === "ENOENT") {
				return mkdirAll(dirname(p)).then(() => mkdirAll(p));
			}

			stat(p, (e, s) => {
				if(e) {
					reject(e);
				} else if(!s.isDirectory()) {
					reject(err);
				} else {
					resolve();
				}
			});
		});
	});
}

function emitFile(filename, data) {
	return new Promise((resolve, reject) => {
		writeFile(filename, data, err => {
			if(err) {
				reject(err);
			} else {
				resolve();
			}
		});
	});
}

export default tsc;
//# sourceMappingURL=rollup-plugin-tsc.es.js.map
